package com.bryceredd.web;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.ArrayList;
import java.util.Stack;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.os.Handler;
import android.os.Looper;

public class ImageLoader {

	FileCache fileCache;

	/*
	 * BEFORE you replace with the another library, let me explain the
	 * requirements of the image loader, and you can see if the new library has
	 * them all: you _must_ be able to:
	 * 
	 * - cache the images on an sd card - scale the images as to avoid crashes
	 * with abnormally large ones (think a 20 mpxl image) - cancel queued image
	 * requests before they have executed - load raw bitmaps (not from
	 * imageviews) - prioritize loading images from the filesystem over url
	 * (optional, but useful)
	 */

	int NUM_THREADS = 3;
	ArrayList<PhotosLoader> threads;
	public PhotosQueue photosQueue = new PhotosQueue();

	public ImageLoader(Context context) {
		initThreads();

		for (PhotosLoader loader : threads)
			loader.setPriority(Thread.NORM_PRIORITY - 1);

		fileCache = new FileCache(context);
	}

	private void initThreads() {
		threads = new ArrayList<PhotosLoader>();
		for (int i = 0; i < NUM_THREADS; i++) {
			threads.add(new PhotosLoader());
		}
	}

	public void loadImage(String url, Handler handler, ImageLoadedResponse response) {
		if (response == null || url == null || url.length() < 2)
			return;

		queuePhoto(url, handler, response);
	}

	private void queuePhoto(final String url, Handler handler, final ImageLoadedResponse response) {

		boolean isInFileCache = fileCache.getFile(url).exists();

		// This ImageView may be used for other images before. So there may be
		// some old tasks in the queue. We need to discard them.
		PhotoToLoad p = new PhotoToLoad(url, response, handler);

		synchronized (photosQueue.photosToLoad) {

			if (isInFileCache)
				photosQueue.photosToLoad.push(p);
			else
				photosQueue.photosToLoad.add(0, p);

			photosQueue.photosToLoad.notify();
		}

		// start thread if it's not started yet
		for (PhotosLoader loader : threads) {
			if (loader.getState() == Thread.State.NEW)
				loader.start();
		}
	}

	private Bitmap getBitmap(String url) {
		File f = fileCache.getFile(url);

		// from SD cache
		Bitmap b = decodeFile(f);
		if (b != null) {
			System.out.println("loading from card " + url);
			return b;
		}

		System.out.println("loading from web " + url);

		// from web
		try {
			URL imageUrl = new URL(url);
			HttpURLConnection conn = (HttpURLConnection) imageUrl.openConnection();
			conn.setConnectTimeout(30000);
			conn.setReadTimeout(30000);
			InputStream is = conn.getInputStream();
			OutputStream os = new FileOutputStream(f);
			Utils.CopyStream(is, os);
			os.close();
			return decodeFile(f);

		} catch (Exception ex) {
			return null;
		}
	}

	// decodes image and scales it to reduce memory consumption
	private Bitmap decodeFile(File f) {
		try {
			// Decode image size
			BitmapFactory.Options o = new BitmapFactory.Options();
			o.inPurgeable = true;
			o.inInputShareable = true;
			o.inJustDecodeBounds = true;
			BitmapFactory.decodeStream(new FileInputStream(f), null, o);

			// The new size we want to scale to
			final int REQUIRED_SIZE = 600;

			// Find the correct scale value. It should be the power of 2.
			int width_tmp = o.outWidth, height_tmp = o.outHeight;
			int scale = 1;
			while (true) {
				if (width_tmp / 2 < REQUIRED_SIZE || height_tmp / 2 < REQUIRED_SIZE)
					break;
				width_tmp /= 2;
				height_tmp /= 2;
				scale *= 2;
			}

			// Decode with inSampleSize
			BitmapFactory.Options o2 = new BitmapFactory.Options();
			o2.inSampleSize = scale;
			Bitmap bitmap = BitmapFactory.decodeStream(new FileInputStream(f), null, o2);

			return bitmap;

		} catch (Exception e) {
		}

		return null;
	}

	public void clearCache() {
		fileCache.clear();
	}

	/* this class is used to ensure the callback happens on the UI thread */
	private class BitmapDisplayer implements Runnable {
		Bitmap bitmap;
		ImageLoadedResponse response;

		public BitmapDisplayer(Bitmap b, ImageLoadedResponse h) {
			bitmap = b;
			response = h;
		}

		public void run() {
			if (bitmap != null) {
				response.imageLoaded(bitmap);
			}else{
				response.failed();
			}

		}
	}
	

	/* a queue for the pending image downloads */
	public class PhotosQueue {
		public Stack<PhotoToLoad> photosToLoad = new Stack<PhotoToLoad>();

		public void Clean(String url) {
			synchronized (photosQueue.photosToLoad) {
				for (int j = 0; j < photosToLoad.size();) {
					if (photosToLoad.get(j).url.equals(url))
						photosToLoad.remove(j);
					else
						++j;
				}
			}
		}
	}

	
	/* an element we can queue */
	private class PhotoToLoad {
		public String url;
		public ImageLoadedResponse response;
		public Handler handler;

		public PhotoToLoad(String u, ImageLoadedResponse r, Handler a) {
			url = u;
			response = r;
			handler = a;
		}
	}

	
	/*
	 * this will create a thread that will continually spin, removing items from
	 * the queue sequentially
	 */
	class PhotosLoader extends Thread {
		public void run() {
			try {
				while (true) {
					
					// thread waits until there are any images to load in the queue
					if (photosQueue.photosToLoad.size() == 0)
						synchronized (photosQueue.photosToLoad) {
							photosQueue.photosToLoad.wait();
						}
					if (photosQueue.photosToLoad.size() != 0) {

						PhotoToLoad photoToLoad;
						synchronized (photosQueue.photosToLoad) {
							if (photosQueue.photosToLoad.size() > 0) {
								try {
									photoToLoad = photosQueue.photosToLoad.pop();
								} catch (Exception e) {
									continue;
								}
							} else
								continue;
						}

						Bitmap bmp = getBitmap(photoToLoad.url);
						BitmapDisplayer bd = new BitmapDisplayer(bmp, photoToLoad.response);
						
						Looper.getMainLooper().getThread().run();
						photoToLoad.handler.post(bd);

					}

					if (Thread.interrupted())
						break;
					
				}
			} catch (InterruptedException e) { }
		}
	}
}
